@@ -1,20 +1,7 @@
+// SPDX-License-Identifier: LGPL-2.1-or-later
 /*
  * Copyright 2012 Cisco Systems, Inc. and/or its affiliates. All rights reserved.
  * Author: Konke Radlow <koradlow@gmail.com>
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation; either version 2.1 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA  02110-1335  USA
  */
 
 #include <errno.h>
@@ -23,11 +10,14 @@
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
-#include <config.h>
 #include <sys/types.h>
 #include <sys/mman.h>
 
+#if defined(__OpenBSD__)
+#include <sys/videoio.h>
+#else
 #include <linux/videodev2.h>
+#endif
 
 #include "../include/libv4l2rds.h"
 
@@ -142,7 +132,7 @@ static uint32_t rds_decode_b(struct rds_private_state *priv_state, struct v4l2_r
 	grp->group_version = (rds_data->msb & 0x08) ? 'B' : 'A';
 
 	/* bit 10 (2 of msb) defines Traffic program Code */
-	traffic_prog = (bool)rds_data->msb & 0x04;
+	traffic_prog = rds_data->msb & 0x04;
 	if (handle->tp != traffic_prog) {
 		handle->tp = traffic_prog;
 		updated_fields |= V4L2_RDS_TP;
@@ -196,6 +186,30 @@ static void rds_decode_d(struct rds_private_state *priv_state, struct v4l2_rds_d
 	grp->data_d_lsb = rds_data->lsb;
 }
 
+/* decodes the RDS radio frequency representation into Hz
+ * @af: 8-bit AF value as transmitted in RDS groups
+ * @is_vhf: boolean value defining  which conversion table to use
+ * @return: frequency in Hz, 0 in case of wrong input values */
+static uint32_t rds_decode_af(uint8_t af, bool is_vhf)
+{
+	uint32_t freq = 0;
+
+	/* AF = 0 => "not to be used"
+	 * AF >= 205 => special meanings */
+	if (af == 0 || af >= 205)
+		return 0;
+
+	/* calculate the AF values in HZ */
+	if (is_vhf)
+		freq = 87500000 + af * 100000;
+	else if (af <= 15)
+		freq = 152000 + af * 9000;
+	else
+		freq = 531000 + af * 9000;
+
+	return freq;
+}
+
 /* compare two rds-groups for equality */
 /* used for decoding RDS-TMC, which has the requirement that the same group
  * is at least received twice before it is accepted */
@@ -219,6 +233,104 @@ static bool rds_compare_group(const struct v4l2_rds_group *a,
 	return true;
 }
 
+/* checks if an entry for the given PI already exists and returns the index
+ * of that entry if so. Else it adds a new entry to the TMC-Tuning table and returns
+ * the index of the new field */
+static int rds_add_tmc_station(struct rds_private_state *priv_state, uint16_t pi)
+{
+	struct v4l2_tmc_tuning *tuning = &priv_state->handle.tmc.tuning;
+	uint8_t index = tuning->index;
+	uint8_t size = tuning->station_cnt;
+
+	/* check if there's an entry for the given PI key */
+	for (int i = 0; i < tuning->station_cnt; i++) {
+		if (tuning->station[i].pi == pi) {
+			return i;
+		}
+	}
+	/* if the the maximum table size is reached, overwrite old
+	 * entries, starting at the oldest one = 0 */
+	tuning->station[index].pi = pi;
+	tuning->index = (index+1 < MAX_TMC_ALT_STATIONS) ? (index+1) : 0;
+	tuning->station_cnt = (size+1 <= MAX_TMC_ALT_STATIONS) ? (size+1) : MAX_TMC_ALT_STATIONS;
+	return index;
+}
+
+/* tries to add new AFs to the relevant entry in the list of RDS-TMC providers */
+static bool rds_add_tmc_af(struct rds_private_state *priv_state)
+{
+	struct v4l2_rds_group *grp = &priv_state->rds_group;
+	struct v4l2_tmc_alt_freq *afi;
+	uint16_t pi_on = grp->data_d_msb << 8 | grp->data_d_lsb;
+	uint8_t variant = grp->data_b_lsb & 0x0f;
+	uint8_t station_index = rds_add_tmc_station(priv_state, pi_on);
+	uint8_t af_index;
+	uint8_t mapped_af_index;
+	uint32_t freq_a = rds_decode_af(grp->data_c_msb, true);
+	uint32_t freq_b = rds_decode_af(grp->data_c_lsb, true);
+
+	afi = &priv_state->handle.tmc.tuning.station[station_index].afi;
+	af_index = afi->af_index;
+	mapped_af_index = afi->mapped_af_index;
+
+	/* specific frequencies */
+	if (variant == 6) {
+		/* compare the new AFs to the stored ones, reset them to 0 if the AFs are
+		 * already known */
+		for (int i = 0; i < afi->af_size; i++) {
+			freq_a = (freq_a == afi->af[i]) ? 0 : freq_a;
+			freq_b = (freq_b == afi->af[i]) ? 0 : freq_b;
+		}
+		/* return early if there is nothing to do */
+		if (freq_a == 0 && freq_b == 0)
+			return false;
+
+		/* add the new AFs if they were previously unknown */
+		if (freq_a != 0) {
+			afi->af[af_index] = freq_a;
+			af_index = (af_index+1 < MAX_TMC_AF_CNT) ? af_index+1 : 0;
+			afi->af_size++;
+		}
+		if (freq_b != 0) {
+			afi->af[af_index] = freq_b;
+			af_index = (af_index+1 < MAX_TMC_AF_CNT) ? af_index+1 : 0;
+			afi->af_size++;
+		}
+		/* update the information in the handle */
+		afi->af_index = af_index;
+		if (afi->af_size >= MAX_TMC_AF_CNT)
+			afi->af_size = MAX_TMC_AF_CNT;
+
+		return true;
+	}
+
+	/* mapped frequency pair */
+	if (variant == 7) {
+		/* check if there's already a frequency mapped to the new tuning
+		 * frequency, update the mapped frequency in this case */
+		for (int i = 0; i < afi->mapped_af_size; i++) {
+			if (freq_a == afi->mapped_af_tuning[i]) {
+				afi->mapped_af[i] = freq_b;
+				return true;
+			}
+		}
+		/* new pair is unknown, add it to the list */
+		if (freq_a != 0 && freq_b != 0) {
+			mapped_af_index = (mapped_af_index+1 >= MAX_TMC_AF_CNT) ? 0 : mapped_af_index + 1;
+			afi->mapped_af[mapped_af_index] = freq_b;
+			afi->mapped_af_tuning[mapped_af_index] = freq_a;
+			afi->mapped_af_size++;
+		}
+		/* update the information in the handle */
+		afi->mapped_af_index = mapped_af_index;
+		if (afi->mapped_af_size >= MAX_TMC_AF_CNT)
+			afi->mapped_af_size = MAX_TMC_AF_CNT;
+
+		return true;
+	}
+	return false;
+}
+
 /* decode additional information of a TMC message into handy representation */
 /* the additional information of TMC messages is submitted in (up to) 4 blocks of
  * 28 bits each, which are to be treated as a consecutive bit-array. This data
@@ -341,6 +453,9 @@ static uint32_t rds_decode_tmc_single_group(struct rds_private_state *priv_state
 	msg.event = ((grp->data_c_msb & 0x07) << 8) | grp->data_c_lsb;
 	/* bits 0-15 of block 4 contain the location */
 	msg.location = (grp->data_d_msb << 8) | grp->data_c_lsb;
+	/* there is no service ID in a single group TMC message, so
+	 * just set it to 0. */
+	msg.sid = 0;
 
 	/* decoding done, store the new message */
 	priv_state->handle.tmc.tmc_msg = msg;
@@ -368,7 +483,7 @@ static uint32_t rds_decode_tmc_multi_group(struct rds_private_state *priv_state)
 	/* bit 15 of block 3 is the first group indicator */
 	if (grp->data_c_msb & 0x80) {
 		/* begine decoding of new message */
-		memset(msg, 0, sizeof(msg));
+		memset(msg, 0, sizeof(*msg));
 		memset(priv_state->optional_tmc, 0, 112*sizeof(bool)); 
 		/* bits 0-3 of block 2 contain continuity index */
 		priv_state->continuity_id = grp->data_b_lsb & 0x07;
@@ -434,16 +549,71 @@ static uint32_t rds_decode_tmc_multi_group(struct rds_private_state *priv_state)
 	return V4L2_RDS_TMC_MG;
 }
 
+/* decode the RDS-TMC tuning information that is contained in type 8A groups
+ * (variants 4 to 9) that announce the presence alternative transmitters 
+ * providing the same RDS-TMC service */
+static uint32_t rds_decode_tmc_tuning(struct rds_private_state *priv_state)
+{
+	struct v4l2_rds_group *group = &priv_state->rds_group;
+	struct v4l2_rds_tmc *tmc = &priv_state->handle.tmc;
+	uint8_t variant_code = group->data_b_lsb & 0x0f;
+	uint16_t pi_on = (group->data_d_msb << 8) | group->data_d_lsb;
+	uint8_t index;
+
+	/* variants 4 and 5 carry the service provider name */
+	if (variant_code >= 4 && variant_code <= 5) {
+		int offset = 4 * (variant_code - 4);
+		tmc->spn[0 + offset] = group->data_c_msb;
+		tmc->spn[1 + offset] = group->data_c_lsb;
+		tmc->spn[2 + offset] = group->data_d_msb;
+		tmc->spn[3 + offset] = group->data_d_lsb;
+
+	/* variant 6 provides specific frequencies for the same RDS-TMC service
+	 * on a network with a different PI code */
+	/* variant 7 provides mapped frequency pair information which should only
+	 * be used if the terminal is tuned to the tuning frequency */
+	} else if (variant_code == 6 || variant_code == 7) {
+		rds_add_tmc_af(priv_state);
+
+	/* variant 8 indicates up to 2 PI codes of adjacent networks carrying 
+	 * the same RDS-TMC service on all transmitters of the network */ 
+	} else if (variant_code == 8) {
+		uint16_t pi_on_2 = (group->data_c_msb << 8) | group->data_c_lsb;
+
+		/* try to add both transmitted PI codes to the table */
+		rds_add_tmc_station(priv_state, pi_on);
+		/* PI = 0 is used as a filler code */
+		if (pi_on_2 != 0)
+			rds_add_tmc_station(priv_state, pi_on_2);
+
+	/* variant 9 provides PI codes of other networks with different system 
+	 * parameters */
+	} else if (variant_code == 9) {
+		index = rds_add_tmc_station(priv_state, pi_on);
+
+		/* bits 0 - 5 contain the service-ID of the ON */
+		tmc->tuning.station[index].sid = group->data_c_lsb & 0x3F;
+		/* bits 6-10 contain the msg parameters of the ON */
+		tmc->tuning.station[index].msg = (group->data_c_msb & 0x03) << 2;
+		tmc->tuning.station[index].msg |= (group->data_c_lsb >> 6) & 0x03;
+		/* bits 11-15 contain the database-ID of the ON */
+		tmc->tuning.station[index].ltn = group->data_c_msb >> 2;
+	}
+
+	return V4L2_RDS_TMC_TUNING;
+}
+
 static bool rds_add_oda(struct rds_private_state *priv_state, struct v4l2_rds_oda oda)
 {
 	struct v4l2_rds *handle = &priv_state->handle;
 
 	/* check if there was already an ODA announced for this group type */
 	for (int i = 0; i < handle->rds_oda.size; i++) {
-		if (handle->rds_oda.oda[i].group_id == oda.group_id)
+		if (handle->rds_oda.oda[i].group_id == oda.group_id) {
 			/* update the AID for this ODA */
 			handle->rds_oda.oda[i].aid = oda.aid;
 			return false;
+		}
 	}
 	/* add the new ODA */
 	if (handle->rds_oda.size >= MAX_ODA_CNT)
@@ -455,20 +625,12 @@ static bool rds_add_oda(struct rds_private_state *priv_state, struct v4l2_rds_od
 /* add a new AF to the list, if it doesn't exist yet */
 static bool rds_add_af_to_list(struct v4l2_rds_af_set *af_set, uint8_t af, bool is_vhf)
 {
-	uint32_t freq = 0;
+	/* convert the frequency to Hz, skip on errors */
+	uint32_t freq = rds_decode_af(af, is_vhf);
 
-	/* AF0 -> "Not to be used" */
-	if (af == 0)
+	if (freq == 0) 
 		return false;
 
-	/* calculate the AF values in HZ */
-	if (is_vhf)
-		freq = 87500000 + af * 100000;
-	else if (freq <= 15)
-		freq = 152000 + af * 9000;
-	else
-		freq = 531000 + af * 9000;
-
 	/* prevent buffer overflows */
 	if (af_set->size >= MAX_AF_CNT || af_set->size >= af_set->announced_af)
 		return false;
@@ -478,7 +640,7 @@ static bool rds_add_af_to_list(struct v4l2_rds_af_set *af_set, uint8_t af, bool
 			return false;
 	}
 	/* it's a new AF, add it to the list */
-	af_set->af[(af_set->size)++] = freq;
+	af_set->af[af_set->size++] = freq;
 	return true;
 }
 
@@ -505,10 +667,15 @@ static bool rds_add_af(struct rds_private_state *priv_state)
 			updated_af = true;
 		c_lsb = 0; /* invalidate */
 	}
-	/* 224..249: announcement of AF count (224=0, 249=25)*/
-	if (c_msb >= 224 && c_msb <= 249)
+	/* 224..249: announcement of AF count (224=0, 249=25) */
+	if (c_msb >= 224 && c_msb <= 249) {
+		if (af_set->announced_af != c_msb - 224) {
+			updated_af = true;
+			af_set->size = 0;
+		}
 		af_set->announced_af = c_msb - 224;
-	/* check if the data represents an AF (for 1 =< val <= 204 the
+	}
+	/* check if the data represents an AF (for 1 <= val <= 204 the
 	 * value represents an AF) */
 	if (c_msb < 205)
 		if (rds_add_af_to_list(af_set, c_msb, true))
@@ -522,6 +689,8 @@ static bool rds_add_af(struct rds_private_state *priv_state)
 	return updated_af;
 }
 
+
+
 /* adds one char of the ps name to temporal storage, the value is validated
  * if it is received twice in a row
  * @pos:	position of the char within the PS name (0..7)
@@ -543,6 +712,44 @@ static bool rds_add_ps(struct rds_private_state *priv_state, uint8_t pos, uint8_
 	return true;
 }
 
+/* checks if an entry for the given PI already exists and returns the index
+ * of that entry if so. Else it adds a new entry to the EON table and returns
+ * the index of the new field */
+static uint8_t rds_add_eon_entry(struct rds_private_state *priv_state, uint16_t pi)
+{
+	struct v4l2_rds *handle = &priv_state->handle;
+	uint8_t index = handle->rds_eon.index;
+	uint8_t size = handle->rds_eon.size;
+
+	/* check if there's an entry for the given PI key */
+	for (int i = 0; i < handle->rds_eon.size; i++) {
+		if (handle->rds_eon.eon[i].pi == pi) {
+			return i;
+		}
+	}
+	/* if the the maximum table size is reached, overwrite old
+	 * entries, starting at the oldest one = 0 */
+	handle->rds_eon.eon[index].pi = pi;
+	handle->rds_eon.eon[index].valid_fields |= V4L2_RDS_PI;
+	handle->rds_eon.index = (index+1 < MAX_EON_CNT) ? (index+1) : 0;
+	handle->rds_eon.size = (size+1 <= MAX_EON_CNT) ? (size+1) : MAX_EON_CNT;
+	return index;
+}
+
+/* checks if an entry for the given PI already exists */
+static bool rds_check_eon_entry(struct rds_private_state *priv_state, uint16_t pi)
+{
+	struct v4l2_rds *handle = &priv_state->handle;
+
+	/* check if there's an entry for the given PI key */
+	for (int i = 0; i <= handle->rds_eon.size; i++) {
+		if (handle->rds_eon.eon[i].pi == pi) {
+			return true;
+		}
+	}
+	return false;
+}
+
 /* group of functions to decode successfully received RDS groups into
  * easily accessible data fields
  *
@@ -590,28 +797,28 @@ static uint32_t rds_decode_group0(struct rds_private_state *priv_state)
 
 	/* bit 2 of block B contains 1 bit of the Decoder Control Information (DI)
 	 * the segment number defines the bit position
-	 * New bits are only accepted the segments arrive in the correct order */
+	 * New bits are only accepted if the segments arrive in the correct order */
 	bool bit2 = grp->data_b_lsb & 0x04;
 	if (segment == 0 || segment == priv_state->next_di_segment) {
 		switch (segment) {
 		case 0:
 			priv_state->new_di = set_bit(priv_state->new_di,
-				V4L2_RDS_FLAG_STEREO, bit2);
+				V4L2_RDS_FLAG_DYNAMIC_PTY, bit2);
 			priv_state->next_di_segment = 1;
 			break;
 		case 1:
 			priv_state->new_di = set_bit(priv_state->new_di,
-				V4L2_RDS_FLAG_ARTIFICIAL_HEAD, bit2);
+				V4L2_RDS_FLAG_COMPRESSED, bit2);
 			priv_state->next_di_segment = 2;
 			break;
 		case 2:
 			priv_state->new_di = set_bit(priv_state->new_di,
-				V4L2_RDS_FLAG_COMPRESSED, bit2);
+				V4L2_RDS_FLAG_ARTIFICIAL_HEAD, bit2);
 			priv_state->next_di_segment = 3;
 			break;
 		case 3:
 			priv_state->new_di = set_bit(priv_state->new_di,
-				V4L2_RDS_FLAG_STATIC_PTY, bit2);
+				V4L2_RDS_FLAG_STEREO, bit2);
 			/* check if the value of DI has changed, and store
 			 * and signal DI update in case */
 			if (handle->di != priv_state->new_di) {
@@ -805,10 +1012,10 @@ static time_t rds_decode_mjd(const struct rds_private_state *priv_state)
 	/* add / subtract the local offset to get the local time.
 	 * The offset is expressed in multiples of half hours */
 	if (priv_state->utc_offset & 0x20) { /* bit 5 indicates -/+ */
-		local_hour -= (offset * 2);
+		local_hour -= offset / 2;
 		local_minute -= (offset % 2) * 30;
 	} else {
-		local_hour += (offset * 2);
+		local_hour += offset / 2;
 		local_minute += (offset % 2) * 30;
 	}
 
@@ -827,14 +1034,14 @@ static time_t rds_decode_mjd(const struct rds_private_state *priv_state)
 	new_time.tm_min = local_minute;
 	new_time.tm_hour = local_hour;
 	new_time.tm_mday = d;
-	new_time.tm_mon = m;
+	new_time.tm_mon = m - 1;
 	new_time.tm_year = y;
 	/* offset (submitted by RDS) that was used to compute the local time,
 	 * expressed in multiples of half hours, bit 5 indicates -/+ */
 	if (priv_state->utc_offset & 0x20)
-		new_time.tm_gmtoff = -2 * offset * 3600;
+		new_time.tm_gmtoff = -offset * 1800;
 	else
-		new_time.tm_gmtoff = 2 * offset * 3600;
+		new_time.tm_gmtoff = offset * 1800;
 
 	/* convert tm struct to time_t value and return it */
 	return mktime(&new_time);
@@ -880,7 +1087,6 @@ static uint32_t rds_decode_group4(struct rds_private_state *priv_state)
 	handle->time = rds_decode_mjd(priv_state);
 	updated_fields |= V4L2_RDS_TIME;
 	handle->valid_fields |= V4L2_RDS_TIME;
-	printf("\nLIB: time_t: %ld", handle->time);
 	return updated_fields;
 }
 
@@ -916,13 +1122,13 @@ static uint32_t rds_decode_group8(struct rds_private_state *priv_state)
 		!(grp->data_b_lsb & V4L2_TMC_TUNING_INFO)) {
 		return rds_decode_tmc_multi_group(priv_state);
 	}
-	/* -> tuning information message, defined for variants 4..9, submitted
-	 * in bits 0-3 of block 2 */
+	/* -> tuning information message, defined for variants 4..9,
+	 * submitted in bits 0-3 of block 2 */
 	tuning_variant = grp->data_b_lsb & 0x0f;
 	if ((grp->data_b_lsb & V4L2_TMC_TUNING_INFO) && tuning_variant >= 4 &&
 		tuning_variant <= 9) {
-		/* TODO: Implement tuning information decoding */
-		return 0;
+		priv_state->handle.valid_fields |= V4L2_RDS_TMC_TUNING;
+		return rds_decode_tmc_tuning(priv_state);
 	}
 
 	return 0;
@@ -981,6 +1187,121 @@ static uint32_t rds_decode_group10(struct rds_private_state *priv_state)
 	return updated_fields;
 }
 
+/* group 14: EON (Enhanced Other Network) information */
+static uint32_t rds_decode_group14(struct rds_private_state* priv_state)
+{
+	struct v4l2_rds *handle = &priv_state->handle;
+	struct v4l2_rds_group *grp = &priv_state->rds_group;
+	struct v4l2_rds_eon *eon_entry;
+	uint32_t updated_fields = 0;
+	uint16_t pi_on;
+	uint16_t lsf_on;
+	uint8_t variant_code;
+	uint8_t eon_index;
+	uint8_t pty_on;
+	bool tp_on, ta_on;
+	bool new_a = false, new_b = false;
+
+	if (grp->group_version != 'A')
+		return 0;
+
+	/* bits 0-3 of group b contain the variant code */
+	variant_code = grp->data_b_lsb & 0x0f;
+
+	/* group d contains the PI code of the ON (Other Network) */
+	pi_on = (grp->data_d_msb << 8) | grp->data_d_lsb;
+
+	/* bit 4 of group b contains the TP status of the ON*/
+	tp_on = grp->data_b_lsb & 0x10;
+	if (rds_check_eon_entry(priv_state, pi_on)) {
+		/* if there's an entry for this PI(ON) update the TP field */
+		eon_index = rds_add_eon_entry(priv_state, pi_on);
+		eon_entry = &handle->rds_eon.eon[eon_index];
+		eon_entry->tp = tp_on;
+		eon_entry->valid_fields |= V4L2_RDS_TP;
+		updated_fields |= V4L2_RDS_EON;
+	}
+
+	/* perform group variant dependent decoding */
+	if ((variant_code >=5 && variant_code <= 11) || variant_code >= 14) {
+		/* 5-9 = mapped FM frequencies -> unsupported
+		 * 10-11 = unallocated
+		 * 14 = PIN(ON) -> unsupported (unused RDS feature)
+		 * 15 = reserved for broadcasters use */
+		return updated_fields;
+	}
+	
+	/* retrieve the EON entry corresponding to the PI(ON) code or add a new
+	 * entry to the table if no entry exists */
+	eon_index = rds_add_eon_entry(priv_state, pi_on);
+	eon_entry = &handle->rds_eon.eon[eon_index];
+
+	/* PS Name */
+	if (variant_code < 4) {
+		eon_entry->ps[variant_code*2] = grp->data_c_msb;
+		eon_entry->ps[variant_code*2+1] = grp->data_c_lsb;
+		eon_entry->valid_fields |= V4L2_RDS_PS;
+		updated_fields |= V4L2_RDS_EON;
+	}
+	/* Alternative frequencies */
+	else if (variant_code == 4) {
+		uint8_t c_msb = grp->data_c_msb;
+		uint8_t c_lsb = grp->data_c_lsb;
+
+		/* 224..249: announcement of AF count (224=0, 249=25) */
+		if (c_msb >= 224 && c_msb <= 249)
+			eon_entry->af.announced_af = c_msb - 224;
+		/* check if the data represents an AF (for 1 =< val <= 204 the
+		 * value represents an AF) */
+		if (c_msb < 205)
+			new_a = rds_add_af_to_list(&eon_entry->af,
+					grp->data_c_msb, true);
+		if (c_lsb < 205)
+			new_b = rds_add_af_to_list(&eon_entry->af,
+					grp->data_c_lsb, true);
+		/* check if one of the frequencies was previously unknown */
+		if (new_a || new_b) {
+			eon_entry->valid_fields |= V4L2_RDS_AF;
+			updated_fields |= V4L2_RDS_EON;
+		}
+	}
+	/* Linkage information */
+	else if (variant_code == 12) {
+		/* group c contains the lsf code */
+		lsf_on = (grp->data_c_msb << 8) | grp->data_c_lsb;
+		/* check if the lsf code is already known */
+		new_a = (eon_entry->lsf == lsf_on);
+		if (new_a) {
+			eon_entry->lsf = lsf_on;
+			eon_entry->valid_fields |= V4L2_RDS_LSF;
+			updated_fields |= V4L2_RDS_EON;
+		}
+	}
+	/* PTY(ON) and TA(ON) */
+	else if (variant_code == 13) {
+		/* bits 15-10 of group c contain the PTY(ON) */
+		pty_on = grp->data_c_msb >> 3;
+		/* bit 0 of group c contains the TA code */
+		ta_on = grp->data_c_lsb & 0x01;
+		/* check if the data is new */
+		new_a = (eon_entry->pty == pty_on);
+		if (new_a) {
+			eon_entry->pty = pty_on;
+			eon_entry->valid_fields |= V4L2_RDS_PTY;
+		}
+		new_b = (eon_entry->ta == ta_on);
+		eon_entry->ta = ta_on;
+		eon_entry->valid_fields |= V4L2_RDS_TA;
+		if (new_a || new_b)
+			updated_fields |= V4L2_RDS_EON;
+	}
+	/* set valid field for EON data, if EON table contains entries */
+	if (handle->rds_eon.size > 0)
+		handle->valid_fields |= V4L2_RDS_EON;
+
+	return updated_fields;
+}
+
 typedef uint32_t (*decode_group_func)(struct rds_private_state *);
 
 /* array of function pointers to contain all group specific decoding functions */
@@ -992,6 +1313,7 @@ static const decode_group_func decode_group[16] = {
 	[4] = rds_decode_group4,
 	[8] = rds_decode_group8,
 	[10] = rds_decode_group10,
+	[14] = rds_decode_group14
 };
 
 static uint32_t rds_decode_group(struct rds_private_state *priv_state)
@@ -1056,7 +1378,8 @@ uint32_t v4l2_rds_add(struct v4l2_rds *handle, struct v4l2_rds_data *rds_data)
 
 	rds_stats->block_cnt++;
 	/* check for corrected / uncorrectable errors in the data */
-	if (rds_data->block & V4L2_RDS_BLOCK_ERROR) {
+	if ((rds_data->block & V4L2_RDS_BLOCK_ERROR) ||
+	    block_id == V4L2_RDS_BLOCK_INVALID) {
 		block_id = -1;
 		rds_stats->block_error_cnt++;
 	} else if (rds_data->block & V4L2_RDS_BLOCK_CORRECTED) {
@@ -1068,7 +1391,7 @@ uint32_t v4l2_rds_add(struct v4l2_rds *handle, struct v4l2_rds_data *rds_data)
 		if (block_id == 0) {
 			*decode_state = RDS_A_RECEIVED;
 			/* begin reception of a new data group, reset raw buffer to 0 */
-			memset(rds_data_raw, 0, sizeof(rds_data_raw));
+			memset(rds_data_raw, 0, sizeof(*rds_data_raw));
 			rds_data_raw[0] = *rds_data;
 		} else {
 			/* ignore block if it is not the first block of a group */
@@ -1251,7 +1574,8 @@ const char *v4l2_rds_get_language_str(const struct v4l2_rds *handle)
 const char *v4l2_rds_get_coverage_str(const struct v4l2_rds *handle)
 {
 	/* bits 8-11 contain the area coverage code */
-	uint8_t coverage = (handle->pi >> 8) & 0X0f;
+	uint8_t coverage = (handle->pi >> 8) & 0x0f;
+	uint8_t nibble = (handle->pi >> 12) & 0x0f;
 	static const char *coverage_lut[16] = {
 		"Local", "International", "National", "Supra-Regional",
 		"Regional 1", "Regional 2", "Regional 3", "Regional 4",
@@ -1259,7 +1583,14 @@ const char *v4l2_rds_get_coverage_str(const struct v4l2_rds *handle)
 		"Regional 9", "Regional 10", "Regional 11", "Regional 12"
 	};
 
-	return coverage_lut[coverage];
+	/*
+	 * Coverage area codes are restricted to the B, D and E PI code
+	 * blocks for RBDS.
+	 */
+	if (!handle->is_rbds ||
+	    (nibble == 0xb || nibble == 0xd || nibble == 0xe))
+		return coverage_lut[coverage];
+	return "Not Available";
 }
 
 const struct v4l2_rds_group *v4l2_rds_get_group(const struct v4l2_rds *handle)
