@@ -12,7 +12,6 @@
    GNU General Public License for more details.
  */
 
-#include <config.h>
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -37,7 +36,7 @@
 #include "keymap.h"
 
 #ifdef HAVE_BPF
-#include "bpf.h"
+#include <bpf/bpf.h>
 #include "bpf_load.h"
 #endif
 
@@ -55,11 +54,11 @@
 
 struct input_keymap_entry_v2 {
 #define KEYMAP_BY_INDEX	(1 << 0)
-	u_int8_t  flags;
-	u_int8_t  len;
-	u_int16_t index;
-	u_int32_t keycode;
-	u_int8_t  scancode[32];
+	uint8_t  flags;
+	uint8_t  len;
+	uint16_t index;
+	uint32_t keycode;
+	uint8_t  scancode[32];
 };
 
 #ifndef input_event_sec
@@ -82,7 +81,7 @@ struct input_keymap_entry_v2 {
 struct keytable_entry {
 	// 64 bit int which can printed with %llx
 	unsigned long long scancode;
-	u_int32_t keycode;
+	uint32_t keycode;
 	struct keytable_entry *next;
 };
 
@@ -382,14 +381,16 @@ static int add_keymap(struct keymap *map, const char *fname)
 
 		protocol = parse_sysfs_protocol(map->protocol, false);
 		if (protocol == SYSFS_INVALID) {
-			struct bpf_protocol *b;
+			if (strcmp(map->protocol, "none")) {
+				struct bpf_protocol *b;
 
-			b = malloc(sizeof(*b));
-			b->name = strdup(map->protocol);
-			b->param = map->param;
-			/* steal param */
-			map->param = NULL;
-			add_bpf_protocol(b);
+				b = malloc(sizeof(*b));
+				b->name = strdup(map->protocol);
+				b->param = map->param;
+				/* steal param */
+				map->param = NULL;
+				add_bpf_protocol(b);
+			}
 		} else {
 			ch_proto |= protocol;
 		}
@@ -1376,7 +1377,7 @@ static int get_input_protocol_version(int fd)
 static void clear_table(int fd)
 {
 	int i, j;
-	u_int32_t codes[2];
+	uint32_t codes[2];
 	struct input_keymap_entry_v2 entry;
 
 	/* Clears old table */
@@ -1463,7 +1464,7 @@ static void display_proto(struct rc_device *rc_dev)
 }
 
 
-static char *get_event_name(struct parse_event *event, u_int16_t code)
+static char *get_event_name(struct parse_event *event, uint16_t code)
 {
 	struct parse_event *p;
 
@@ -1647,14 +1648,14 @@ static void display_table_v2(struct rc_device *rc_dev, int fd)
 		if (ioctl(fd, EVIOCGKEYCODE_V2, &entry) == -1)
 			break;
 
-		if (entry.len == sizeof(u_int32_t)) {
-			u_int32_t temp;
+		if (entry.len == sizeof(uint32_t)) {
+			uint32_t temp;
 
 			memcpy(&temp, entry.scancode, sizeof(temp));
 
 			scancode = temp;
-		} else if (entry.len == sizeof(u_int64_t)) {
-			u_int64_t temp;
+		} else if (entry.len == sizeof(uint64_t)) {
+			uint64_t temp;
 
 			memcpy(&temp, entry.scancode, sizeof(temp));
 
@@ -1701,7 +1702,7 @@ static int get_rate(int fd, unsigned int *delay, unsigned int *period)
 	}
 	*delay = rep[0];
 	*period = rep[1];
-	printf(_("Repeat delay = %d ms, repeat period = %d ms\n"), *delay, *period);
+	printf(_("Repeat delay: %d ms, repeat period: %d ms\n"), *delay, *period);
 	return 0;
 }
 
@@ -1751,7 +1752,7 @@ static bool attach_bpf(const char *lirc_name, const char *bpf_prog, struct proto
 	struct rlimit rl;
 	int fd, ret;
 
-	fd = open(lirc_name, O_RDONLY);
+	fd = open(lirc_name, O_RDWR);
 	if (fd == -1) {
 		perror(lirc_name);
 		return false;
@@ -1842,7 +1843,7 @@ static void clear_bpf(const char *lirc_name)
 	unsigned int features, i;
 	int ret, prog_fd, fd;
 
-	fd = open(lirc_name, O_RDONLY);
+	fd = open(lirc_name, O_RDWR);
 	if (fd == -1) {
 		perror(lirc_name);
 		return;
@@ -2105,11 +2106,6 @@ int main(int argc, char *argv[])
 			}
 			add_keymap(map, fname);
 			free_keymap(map);
-			if (!keytable) {
-				fprintf(stderr, _("Empty keymap %s\n"), fname);
-				free(fname);
-				return -1;
-			}
 			free(fname);
 			clear = 1;
 			matches++;
